from typing import List, Tuple

import pandas as pd

from .DataPoint import DataPoint


def trainTestSplit(
    df: pd.DataFrame, ratio: float = 0.8
) -> Tuple[List[DataPoint], List[DataPoint]]:
    data_points = []
    cols = df.columns
    inputs = []
    for i in range(len(df[cols[0]])):
        inputs = [df[cols[j]][i] for j in range(len(cols) - 1)]
        output = df[cols[-1]][i]
        expected_outputs = [0.0] * (int(df[cols[-1]].max()) + 1)
        expected_outputs[int(output)] = 1.0
        data_points.append(DataPoint(inputs, expected_outputs))

    train_data = data_points[: int(ratio * len(data_points))]
    test_data = data_points[int(ratio * len(data_points)) :]

    return train_data, test_data



from typing import Any, Dict

import numpy as np
import pandas as pd


class dataCleaner:
    """A tool to clean datasets for better training of models"""

    def __init__(self, train: pd.DataFrame, test: pd.DataFrame) -> None:
        """
        Parameters
        ------------
        train: pd.DataFrame
            DataFrame that will be used to train the model

        test: pd.DataFrame
            DataFrame that will be used to evaluate the trained model
        """

        self.train: pd.DataFrame = train.copy()
        self.test: pd.DataFrame = test.copy()
        self.variables = train.columns.tolist()

    def show_duplicate_observations(
        self,
    ) -> Dict[str, pd.DataFrame | pd.Series]:
        """Returns a dict of duplicate rows between train and test"""

        train_duplicates = self.train.duplicated()
        test_duplicates = self.test.duplicated()

        return {
            "train": self.train[train_duplicates].copy(),
            "test": self.test[test_duplicates.copy()],
        }

    def remove_duplicate_observations(self) -> None:
        """Deletes duplicate rows"""

        self.train.drop_duplicates(inplace=True)
        self.test.drop_duplicates(inplace=True)

    def show_missing_values(self) -> Dict[str, int]:
        """Returns a dict containing the number of missing values in columns"""

        train_missing_cells = self.train.isna().sum()
        test_missing_cells = self.test.isna().sum()

        return {"train": train_missing_cells, "test": test_missing_cells}

    def fix_missing_values(
        self, feature: str = "", strategy: str = "Delete", n_neighbours: int = 3
    ) -> None:
        """
        Fixes the empty values in the datasets by the strategy of your choosing:
            KNN: Uses K Nearest Neigbours method to fill the empty cells
            Mean: Uses the mean value of the column to fill the empty cells
            Mode: Uses the most frequent element in the column to fill the empty cells
            Median: Uses the median value of the column to fill the empty cells
            Delete: Deletes the rows with empty values. Default

        --------------

        feature: string
            Column name to be fixed
        strategy: string
            Strategy of fixing
        n_neighbours: integer
            Number of closest neighbours to check in KNN strategy
        """

        train_notna_idx = None
        test_na_idx = None
        train_na_idx = None

        if strategy != "KNN":
            if not feature:
                raise ValueError("Your strategy requires a column name to be provided")

            train_notna_idx = self.train.loc[:, feature].notna().values
            test_na_idx = self.train.loc[:, feature].isna().values
            train_na_idx = self.train.loc[:, feature].isna().values

        if strategy == "Delete":
            self.train = self.train.drop(columns=feature)
            self.test = self.test.drop(columns=feature)

        elif strategy == "Mean":
            mean = self.train.loc[train_notna_idx, feature].mean()
            print("Mean: ", mean)
            self.train.loc[train_na_idx, feature] = mean
            self.test.loc[test_na_idx, feature] = mean

        elif strategy == "Mode":
            mode = self.train.loc[train_notna_idx, feature].mode()[0]
            print("Mode: ", mode)
            self.train.loc[train_na_idx, feature] = mode
            self.test.loc[test_na_idx, feature] = mode

        elif strategy == "Median":
            median = self.train.loc[train_notna_idx, feature].median()
            print("Mode: ", median)
            self.train.loc[train_na_idx, feature] = median
            self.test.loc[test_na_idx, feature] = median

        else:
            raise ValueError("Are you sure you chose the strategy correctly")

    def outlier_detection(
        self,
        feature: str = "",
        strategy: str = "inter_quartile_range",
        n_estimators: int = 50,
        contamination: float | str = "auto",
        max_samples: float | str = "auto",
    ) -> Dict[str, Any]:
        """
        Finds the outliers

        Args:
            feature (str, optional): column name. Defaults to "".
            strategy (str, optional): strategy for outlier_detection. Defaults to "inter_quartile_range".
            n_estimators (int, optional): Defaults to 50.
            contamination (float | int, optional): Defaults to 0.1.
        """

        train_outliers = None
        test_outliers = None

        if strategy == "inter_quartile_range":
            Q1 = self.train.loc[:, feature].quantile(0.25)
            Q3 = self.train.loc[:, feature].quantile(0.75)

            IQR = Q3 - Q1

            lower_limit = Q1 - 1.5 * IQR
            upper_limit = Q3 + 1.5 * IQR

            print("Upper Limit: ", upper_limit, "\nLower Limit: ", lower_limit)

            train_out = np.array(
                (self.train[feature] < lower_limit)
                & (self.train[feature] > upper_limit)
            )

            test_out = np.array(
                (self.test[feature] < lower_limit) & (self.test[feature] > upper_limit)
            )

            train_outliers = self.train.loc[train_out]
            test_outliers = self.test.loc[test_out]

        else:
            raise ValueError("Wrong method choice")

        return {"train": train_outliers, "test": test_outliers}



from dataclasses import dataclass, field
from typing import Callable, Dict

import numpy as np


@dataclass
class Layer:
    num_nodes_in: int
    num_nodes_out: int
    activation_function: str
    error_func: str = "mse"
    weights: np.ndarray = field(init=False)
    cost_gradient_w: np.ndarray = field(init=False)
    biases: np.ndarray = field(init=False)
    cost_gradient_b: np.ndarray = field(init=False)
    nodeCost: Callable[[np.ndarray, np.ndarray], float] = field(init=False)
    _activationFunction: Callable[[np.ndarray], np.ndarray] = field(init=False)
    _activationFunctionDerivative: Callable[[np.ndarray], np.ndarray] = field(
        init=False
    )

    def __post_init__(self):
        self.weights = np.random.randn(self.num_nodes_in, self.num_nodes_out) / np.sqrt(
            self.num_nodes_in
        )
        self.biases = np.random.randn(self.num_nodes_out)
        self.cost_gradient_w = np.zeros((self.num_nodes_in, self.num_nodes_out))
        self.cost_gradient_b = np.zeros(self.num_nodes_out)

        cost_map: Dict[str, Callable[[np.ndarray, np.ndarray], float]] = {
            "mse": self._mse,
            "cross_entropy": self._cross_entropy,
        }
        activation_map: Dict[str, Callable[[np.ndarray], np.ndarray]] = {
            "sigmoid": self._sigmoid,
            "relu": self._relu,
            "leaky_relu": self._leaky_relu,
            "tanh": self._tanh,
            "softmax": self._softmax,
        }
        derivative_map: Dict[str, Callable[[np.ndarray], np.ndarray]] = {
            "sigmoid": self._sigmoid_derivative,
            "relu": self._relu_derivative,
            "leaky_relu": self._leaky_relu_derivative,
            "tanh": self._tanh_derivative,
            "softmax": self._softmax_derivative,
        }

        self.nodeCost = cost_map[self.error_func]
        self._activationFunction = activation_map[self.activation_function]
        self._activationFunctionDerivative = derivative_map[self.activation_function]

    def calculateOutputs(self, inputs: np.ndarray) -> np.ndarray:
        weighted_input = np.dot(inputs, self.weights) + self.biases
        return self._activationFunction(weighted_input)

    def applyGradient(self, learn_rate: float):
        self.weights -= learn_rate * self.cost_gradient_w
        self.biases -= learn_rate * self.cost_gradient_b

    @staticmethod
    def _mse(y_true: float, y_pred: float) -> float:
        return (y_true - y_pred) ** 2

    @staticmethod
    def _cross_entropy(y_true: float, y_pred: float) -> float:
        return -np.sum(y_true * np.log(y_pred + 1e-9))

    @staticmethod
    def _sigmoid(z: np.ndarray) -> np.ndarray:
        return 1.0 / (1.0 + np.exp(-z))

    @staticmethod
    def _sigmoid_derivative(z: np.ndarray) -> np.ndarray:
        s = Layer._sigmoid(z)
        return s * (1 - s)

    @staticmethod
    def _relu(z: np.ndarray) -> np.ndarray:
        return np.maximum(0, z)

    @staticmethod
    def _relu_derivative(z: np.ndarray) -> np.ndarray:
        return (z > 0).astype(float)

    @staticmethod
    def _leaky_relu(z: np.ndarray) -> np.ndarray:
        return np.where(z > 0, z, 0.01 * z)

    @staticmethod
    def _leaky_relu_derivative(z: np.ndarray) -> np.ndarray:
        return np.where(z > 0, 1, 0.01)

    @staticmethod
    def _tanh(z: np.ndarray) -> np.ndarray:
        return np.tanh(z)

    @staticmethod
    def _tanh_derivative(z: np.ndarray) -> np.ndarray:
        return 1.0 - np.tanh(z) ** 2

    @staticmethod
    def _softmax(z: np.ndarray) -> np.ndarray:
        e_z = np.exp(z - np.max(z))
        return e_z / e_z.sum(axis=0, keepdims=True)

    @staticmethod
    def _softmax_derivative(z: np.ndarray) -> np.ndarray:
        s = Layer._softmax(z)
        return s * (
            1 - s
        )  # Note: Typically used with the Jacobian matrix for accurate calculations



import pickle
from typing import List

import numpy as np

from ..Utils.DataPoint import DataPoint
from ..Utils.progress_bar import ProgressBar
from .Layer import Layer


class NeuralNetwork:
    def __init__(self, layers: List[Layer]) -> None:
        self.layers = layers
        self.velocity_w = [np.zeros_like(layer.weights) for layer in layers]
        self.velocity_b = [np.zeros_like(layer.biases) for layer in layers]

    def calculateOutputs(self, inputs: np.ndarray) -> np.ndarray:
        for layer in self.layers:
            inputs = layer.calculateOutputs(inputs)
        return inputs

    def _classify(self, inputs: np.ndarray) -> np.intp:
        outputs = self.calculateOutputs(inputs)
        return np.argmax(outputs)

    def classify(self, inputs: List[DataPoint]) -> List[np.intp]:
        preds = []
        for point in inputs:
            pred = self._classify(point.inputs)
            preds.append(pred)
        return preds

    def _cost(self, data_point: DataPoint) -> float:
        outputs = self.calculateOutputs(np.array(data_point.inputs))
        output_layer = self.layers[-1]

        cost = 0.0
        for out_node in range(len(outputs)):
            cost += output_layer.nodeCost(
                outputs[out_node], data_point.expected_outputs[out_node]
            )
        return cost

    def cost(self, data_points: List[DataPoint]) -> float:
        total_cost = 0.0
        for data_point in data_points:
            total_cost += self._cost(data_point)
        return total_cost / len(data_points)

    def learn(
        self,
        training_data: List[DataPoint],
        learning_rate: float,
        batch_size: int,
        momentum: float = 0.9,
    ) -> float:
        batches = [
            training_data[k : k + batch_size]
            for k in range(0, len(training_data), batch_size)
        ]

        for batch in batches:
            nabla_w = [np.zeros_like(layer.weights) for layer in self.layers]
            nabla_b = [np.zeros_like(layer.biases) for layer in self.layers]

            for data_point in batch:
                activation = np.array(data_point.inputs)
                activations = [activation]
                zs = []
                for layer in self.layers:
                    z = np.dot(activation, layer.weights) + layer.biases
                    zs.append(z)
                    activation = layer._activationFunction(z)
                    activations.append(activation)

                delta = (
                    activations[-1] - np.array(data_point.expected_outputs)
                ) * self.layers[-1]._activationFunctionDerivative(zs[-1])
                nabla_b[-1] += delta
                nabla_w[-1] += np.outer(activations[-2], delta)

                for layer_idx in range(2, len(self.layers) + 1):
                    z = zs[-layer_idx]
                    sp = self.layers[-layer_idx]._activationFunctionDerivative(z)
                    delta = np.dot(delta, self.layers[-layer_idx + 1].weights.T) * sp
                    nabla_b[-layer_idx] += delta
                    nabla_w[-layer_idx] += np.outer(activations[-layer_idx - 1], delta)

            for i, layer in enumerate(self.layers):
                self.velocity_w[i] = (
                    momentum * self.velocity_w[i]
                    - (learning_rate / len(batch)) * nabla_w[i]
                )
                self.velocity_b[i] = (
                    momentum * self.velocity_b[i]
                    - (learning_rate / len(batch)) * nabla_b[i]
                )
                layer.weights += self.velocity_w[i]
                layer.biases += self.velocity_b[i]

        return self.cost(training_data)

    def train(
        self,
        iterations: int,
        data_points: List[DataPoint],
        learning_rate: float,
        batch_size: int,
        decay: float = 1e-7,
        momentum: float = 0.9,
    ) -> List[float]:
        cost = None
        costs = []
        progress = ProgressBar(total=iterations, program_name="training.", cost=cost)
        for _ in range(iterations):
            # learning_rate = learning_rate * (1.0 / 1.0 + decay * iterations)
            cost = self.learn(data_points, learning_rate, batch_size, momentum)
            progress.increment(cost=cost)
            costs.append(cost)
        return costs

    def save_model(self, file_path: str):
        with open(file_path, "wb") as file:
            pickle.dump(self, file)

    @classmethod
    def load_model(cls, file_path: str) -> "NeuralNetwork":
        with open(file_path, "rb") as file:
            return pickle.load(file)
